//
// C++ Implementation: bot
//
// Description: 
//
// Contains code of the base bot class
//
// Author:  Rick <rickhelmus@gmail.com>
//
//
//

#include "../cube.h"

dvector bots;

extern int triggertime;
extern itemstat itemstats[];

const vec g_vecZero = {0,0,0};

//Bot class begin

void CBot::CreateInit(dynent *d, char *name, char *skill)
{
    d->bIsBot = true;
    m_pMyEnt = d;
    m_iLastBotUpdate = 0;
    m_bSendC2SInit = false;

    if (name && *name)
        strn0cpy(d->name, name, 16);
    else
        strn0cpy(d->name, BotManager.GetBotName(), 16);

    if (skill && *skill && strcasecmp(skill, "default"))
    {
        if (!strcasecmp(skill, "best"))
            m_sSkillNr = 0;
        else if (!strcasecmp(skill, "good"))
            m_sSkillNr = 1;
        else if (!strcasecmp(skill, "medium"))
            m_sSkillNr = 2;
        else if (!strcasecmp(skill, "worse"))
            m_sSkillNr = 3;
        else if (!strcasecmp(skill, "bad"))
            m_sSkillNr = 4;
        else if (!strcasecmp(skill, "random"))
            m_sSkillNr = RandomLong(0, 4);
        else
        {
            conoutf("Wrong skill specified. Should be best, good, medium, "
                    "worse, bad, default or random");
            conoutf("Using default skill...");
            m_sSkillNr = BotManager.m_sBotSkill;
        }
    }
    else // No skill specified, use default
        m_sSkillNr = BotManager.m_sBotSkill;
          
    m_pBotSkill = &BotManager.m_BotSkills[m_sSkillNr];

    // Sync waypoints
    SyncWaypoints();

    Spawn();
}

CBot::~CBot()
{
     // Delete all waypoints
     loopi(MAX_MAP_GRIDS)
     {
          loopj(MAX_MAP_GRIDS)
          {
               while (m_WaypointList[i][j].Empty() == false)
                    delete m_WaypointList[i][j].Pop();
          }
     }
}

void CBot::Spawn()
{
     // Init all bot variabeles
     m_pMyEnt->eyeheight = 3.2f;
     m_pMyEnt->aboveeye = 0.7f;
     m_pMyEnt->radius = 1.1f;
         
     spawnplayer(m_pMyEnt);
     
     m_pMyEnt->targetyaw = m_pMyEnt->yaw = m_pMyEnt->targetpitch = m_pMyEnt->pitch = 0.0f;
     m_pMyEnt->move = 0;
     m_pMyEnt->enemy = NULL;
     m_pMyEnt->maxspeed = 22.0f;
     m_pMyEnt->health = 100;
     m_pMyEnt->armour = 50;
     m_pMyEnt->pitch = 0;
     m_pMyEnt->roll = 0;
     m_pMyEnt->state = CS_ALIVE;
     m_pMyEnt->anger = 0;
     m_pMyEnt->pBot = this;     
         
     m_eCurrentBotState = STATE_NORMAL;
     m_iShootDelay = m_iChangeWeaponDelay = 0;
     m_iCheckEnvDelay = 0;
     m_vPrevOrigin = g_vecZero;
     m_iStuckCheckDelay = lastmillis + 250;
     m_bStuck = false;
     m_iStuckTime = 0;
     m_iStrafeCheckDelay = 0;
     m_iMoveDir = DIR_NONE;
     m_iMoveTime = 0;
     m_bAddMoveDir = false;

     m_pPrevEnemy = NULL;
     m_iEnemySearchDelay = 0;
     m_iSawEnemyTime = 0;
     m_bCombatJump = false;
     m_iCombatJumpDelay = 0;
     m_bShootAtFeet = (RandomLong(1, 100) <= m_pBotSkill->sShootAtFeetWithRLPercent);
     m_iHuntDelay = 0;
     m_vHuntLocation = m_vPrevHuntLocation = g_vecZero;
     m_pHuntTarget = NULL;
     m_fPrevHuntDist = 0.0f;
     m_iHuntLastTurnLessTime = m_iHuntPlayerUpdateTime = m_iHuntPauseTime = 0;
     m_iLookAroundDelay = m_iLookAroundTime = m_iLookAroundUpdateTime = 0;
     m_fLookAroundYaw = 0.0f;
     m_bLookLeft = false;
        
     m_iLastJumpPad = 0;
     m_pTargetEnt = NULL;
     while(!m_UnreachableEnts.Empty()) delete m_UnreachableEnts.Pop();
     m_iCheckTeleporterDelay = m_iCheckJumppadsDelay = 0;
     m_iCheckEntsDelay = 0;
     m_iCheckTriggersDelay = 0;
     m_iLookForWaypointTime = 0;
        
     m_iAimDelay = 0;
     m_fYawToTurn = m_fPitchToTurn = 0.0f;
     
     m_vGoal = m_vSecondaryGoal = m_vWaterGoal = g_vecZero;
     
     ResetWaypointVars();
}

void CBot::Think()
{
     if (intermission)
          return;
             
     // Bot is dead?
     if (m_pMyEnt->state == CS_DEAD)
     {
          if(lastmillis-m_pMyEnt->lastaction<1200)
          {
               m_pMyEnt->move = 0;
               moveplayer(m_pMyEnt, 1, false);
          }
          else if (!m_arena)
               Spawn();
     
          SendBotInfo();
          return;
     }
     
     CheckItemPickup();
     
     TLinkedList<unreachable_ent_s*>::node_s *p = m_UnreachableEnts.GetFirst(), *tmp;
     while(p)
     {
          if ((lastmillis - p->Entry->time) > 2000)
          {
               tmp = p;
               p = p->next;
               delete tmp->Entry;               
               m_UnreachableEnts.DeleteNode(tmp);
               continue;
          }
          p = p->next;
     }
     
     if (!BotManager.IdleBots())
     {
          MainAI();
     }
     else
     {
          ResetMoveSpeed();
     }
          
                  
     // Aim to ideal yaw and pitch
     AimToIdeal();
     
     // Store current location, to see if the bot is stuck
     m_vPrevOrigin = m_pMyEnt->o;
     
     // Don't check for stuck if the bot doesn't want to move
     if (!m_pMyEnt->move && !m_pMyEnt->strafe)
          m_iStuckCheckDelay = max(m_iStuckCheckDelay, (lastmillis+100));
          
     // Move the bot
     moveplayer(m_pMyEnt, 1, false);
     
     // Update bot info on all clients
     SendBotInfo();
}

void CBot::AimToVec(const vec &o)
{
    vec aim = o;
    vsub(aim, m_pMyEnt->o);
    m_pMyEnt->targetpitch = atan2(aim.z, sqrt(aim.x * aim.x + aim.y * aim.y))*180/PI;
    m_pMyEnt->targetyaw = -(float)atan2(o.x - m_pMyEnt->o.x, o.y - m_pMyEnt->o.y)/PI*180+180;
}

void CBot::AimToIdeal()
{
     if (m_iAimDelay > lastmillis)
          return;

     float MaxXSpeed = RandomFloat(m_pBotSkill->flMinAimXSpeed,
                                   m_pBotSkill->flMaxAimXSpeed);
     float MaxYSpeed = RandomFloat(m_pBotSkill->flMinAimYSpeed,
                                   m_pBotSkill->flMaxAimYSpeed);
     float XOffset = RandomFloat(m_pBotSkill->flMinAimXOffset,
                                 m_pBotSkill->flMaxAimXOffset);
     float YOffset = RandomFloat(m_pBotSkill->flMinAimYOffset,
                                 m_pBotSkill->flMaxAimYOffset);
     float RealXOffset, RealYOffset;
     float AimXSpeed = MaxXSpeed, AimYSpeed = MaxYSpeed;
     float XDiff = fabs(m_pMyEnt->targetpitch - m_pMyEnt->pitch);
     float YDiff = fabs(m_pMyEnt->targetyaw - m_pMyEnt->yaw);

     // How higher the diff, how higher the offsets and aim speed
          
     if (XOffset)
     {
          if (RandomLong(0, 1))
               RealXOffset = XDiff * (XOffset / 100.0f);
          else
               RealXOffset = -(XDiff * (XOffset / 100.0f));
     }
     else
          RealXOffset = 0.0f;

     if (YOffset)
     {
          if (RandomLong(0, 1))
               RealYOffset = YDiff * (YOffset / 100.0f);
          else
               RealYOffset = -(YDiff * (YOffset / 100.0f));
     }
     else
          RealYOffset = 0.0f;

          
     if (XDiff >= 1.0f)
          AimXSpeed = (AimXSpeed * (XDiff / 80.0f)) + (AimXSpeed * 0.25f);
     else
          AimXSpeed *= 0.01f;

     if (YDiff >= 1.0f)
          AimYSpeed = (AimYSpeed * (YDiff / 70.0f)) + (AimYSpeed * 0.25f);
     else
          AimYSpeed *= 0.015f;

     m_fPitchToTurn = fabs((m_pMyEnt->targetpitch + RealXOffset) - m_pMyEnt->pitch);
     m_fYawToTurn = fabs((m_pMyEnt->targetyaw + RealYOffset) - m_pMyEnt->yaw);
     
     float flIdealPitch = ChangeAngle(AimXSpeed, m_pMyEnt->targetpitch + RealXOffset, m_pMyEnt->pitch);
     float flIdealYaw = ChangeAngle(AimYSpeed, m_pMyEnt->targetyaw + RealYOffset, m_pMyEnt->yaw);
     
//     m_pMyEnt->pitch = WrapXAngle(m_pMyEnt->targetpitch); // Uncomment for instant aiming
//     m_pMyEnt->yaw = WrapYZAngle(m_pMyEnt->targetyaw);
                 
     m_pMyEnt->pitch = WrapXAngle(flIdealPitch);
     m_pMyEnt->yaw = WrapYZAngle(flIdealYaw);
}     

// Function code by Botman
float CBot::ChangeAngle(float speed, float ideal, float current)
{
     float current_180;  // current +/- 180 degrees
     float diff;

     // find the difference in the current and ideal angle
     diff = fabs(current - ideal);
    
     // speed that we can turn during this frame...
     speed = speed * (float(BotManager.m_iFrameTime)/1000.0f);

     // check if difference is less than the max degrees per turn
     if (diff < speed)
          speed = diff;  // just need to turn a little bit (less than max)

     // check if the bot is already facing the idealpitch direction...
     if (diff <= 1.0f)
          return ideal;
          
     if ((current >= 180) && (ideal >= 180))
     {
          if (current > ideal)
               current -= speed;
          else
               current += speed;
     }
     else if ((current >= 180) && (ideal < 180))
     {
          current_180 = current - 180;

          if (current_180 > ideal)
               current += speed;
          else
               current -= speed;
     }
     else if ((current < 180) && (ideal >= 180))
     {
          current_180 = current + 180;
          
          if (current_180 > ideal)
               current += speed;
          else
               current -= speed;
     }
     else  // (current < 180) && (ideal < 180)
     {
          if (current > ideal)
               current -= speed;
          else
               current += speed;
     }


     return current;     
}

float CBot::GetDistance(const vec &o)
{
     vdist(flDist, vecto, m_pMyEnt->o, o);
     return flDist;
}

float CBot::GetDistance(const vec &v1, const vec &v2)
{
     vdist(flDist, vecto, v1, v2);
     return flDist;
}

float CBot::GetDistance(entity *e)
{
     vec v = { e->x, e->y, e->z };
     vdist(flDist, vecto, m_pMyEnt->o, v);
     return flDist;
}

bool CBot::SelectGun(int Gun)
{
     if (m_pMyEnt->ammo[Gun])
     {
          if (m_pMyEnt->gunselect != Gun)
               botplaysound(S_WEAPLOAD, m_pMyEnt);
               
          m_pMyEnt->gunselect = Gun;
          return true;
     }
     return false;
}

bool CBot::IsVisible(entity *e, bool CheckPlayers)
{
     vec v = { e->x, e->y, e->z };
     return ::IsVisible(m_pMyEnt->o, v, (CheckPlayers) ? m_pMyEnt : NULL);
}

bool CBot::IsVisible(vec o, int Dir, float flDist, bool CheckPlayers, float *pEndDist)
{
     static vec angles, end, forward, right, up;
     static traceresult_s tr;
        
     end = o;
     angles = GetViewAngles();
     angles.x = 0;
          
     if (Dir & UP)
          angles.x = WrapXAngle(angles.x + 45.0f);
     else if (Dir & DOWN)
          angles.x = WrapXAngle(angles.x - 45.0f);
          
     if ((Dir & FORWARD) || (Dir & BACKWARD))
     {
          if (Dir & BACKWARD)
               angles.y = WrapYZAngle(angles.y + 180.0f);
          
          if (Dir & LEFT)
          {
               if (Dir & FORWARD)
                    angles.y = WrapYZAngle(angles.y - 45.0f);
               else
                    angles.y = WrapYZAngle(angles.y + 45.0f);
          }
          else if (Dir & RIGHT)
          {
               if (Dir & FORWARD)
                    angles.y = WrapYZAngle(angles.y + 45.0f);
               else
                    angles.y = WrapYZAngle(angles.y - 45.0f);
          }
     }
     else if (Dir & LEFT)
          angles.y = WrapYZAngle(angles.y - 90.0f);
     else if (Dir & RIGHT)
          angles.y = WrapYZAngle(angles.y + 90.0f);
     else if (Dir & UP)
          angles.x = WrapXAngle(angles.x + 90.0f);
     else if (Dir & DOWN)
          angles.x = WrapXAngle(angles.x - 90.0f);
               
     AnglesToVectors(angles, forward, right, up);
     
     vmul(forward, flDist);
     vadd(end, forward);
         
     TraceLine(o, end, m_pMyEnt, CheckPlayers, &tr);
     
     if (pEndDist)
          *pEndDist = GetDistance(o, tr.end);
          
     return !tr.collided;
}

void CBot::SetMoveDir()
{
     if (m_iMoveDir & FORWARD)
          m_pMyEnt->move = 1;
     else if (m_iMoveDir & BACKWARD)
          m_pMyEnt->move = -1;
     else if (!m_bAddMoveDir)
          m_pMyEnt->move = 0;

     if (m_iMoveDir & LEFT)
          m_pMyEnt->strafe = 1;
     else if (m_iMoveDir & RIGHT)
          m_pMyEnt->strafe = -1;
     else if (!m_bAddMoveDir)
          m_pMyEnt->strafe = 0;
}

// Bot class end
